#include <cstdlib>

#include <iostream>
#include <vector>
#include <map>

#include "turing_cartesiangrid.hh"
#include "turing_vtkwriter.hh"


/**
 * Executes the simulation.
 */
int main ()
{
  try
    {
      // a map for collecting the options
      std::map<std::string,std::string> options;

      // set default parameters
      options["L"] = "100";     // diameter of \Omega in each directio
      options["D"] = "128";  // diameter in cells (D = 1<<7 = 2^7 = 128, -> bitshift)
      options["o"] = "1e-1";  // time step size for data output
      options["T"] = "250";  // maximum time

      // parse command line parameters
      for (int i = 1; i < argc; i++)
	{
	  std::string key = argv[i];
	  const std::string::size_type sub = key.find("=");
	  if (sub == std::string::npos)
	    throw NBody_Exception("Failed to parse option " + key);
	  std::string val = key.substr(sub+1);
	  key.erase(sub);
	  options[key] = val;
	}

      // read parameters
      const double L = std::stod(options["L"]);
      const std::size_t D = std::stoul(options["D"]);
      const double o = std::stod(options["o"]);
      const double T = std::stod(options["T"]);

      // create grid
      Turing_CartesianGrid grid({D,D},L/D);
      std::clog << "Created a cartesian grid with " << grid.size() << " cells"
		<< std::endl;

      // allocate memory for data
      std::vector<double> a(grid.size());
      std::vector<double> b(grid.size());

      // create VTK writers
      Turing_VTKWriter vtkWriter_a("turing_a",grid,a);
      Turing_VTKWriter vtkWriter_b("turing_b",grid,b);

      // set initial values
      srand48(42);
      for (unsigned int i = 0; i < grid.size(); i++)
	{
	  a[i] = 1e-2 * drand48();
	  b[i] = 0.0;
	}

      // output initial values to VTK file
      vtkWriter_a.write();
      vtkWriter_b.write();

      // prepare time loop
      double t = 0.0;
      double v = o;

      // time loop
      while (t < T)
	{
	  // do one time step
	  // TODO: choose time step size appropriately
    
	  // TODO: do one time step
	  std::cout << "time step -- from t=" << t << ", dt: " << dt << " to "
		    << t+dt << std::endl;
	  // ...

	  // data output, controlled by time step size vt
	  if (v <= t+dt)
	    {
	      std::cout << "write vtk file -- time: " << t+dt << std::endl;

	      // output solution at time t+dt to VTK file
	      vtkWriter_a.write();
	      vtkWriter_b.write();

	      // prepare next data output
	      v += o;
	    }

	  // prepare next time step
	  t += dt;
	}

      // output solution at final time to VTK file
      vtkWriter_a.write();
      vtkWriter_b.write();
    }
  catch (std::exception& e)
    {
      std::cout << "ERROR: " << e.what() << std::endl;
    }

  // exit without an error
  return 0;
}
