#ifndef TURING_CARTESIANGRID_HH
#define TURING_CARTESIANGRID_HH

#include <cassert>

#include <vector>
#include <array>
#include <numeric>
#include <utility>


/**
 * A class for generating informations about a Cartesian grid
 * which can be used in a cell centered finite volume scheme.
 * TODO: The class currently is limited to Cartesian grids in 2D.
 *       Generalize it to Cartesian grids in 1D/2D/3D in a future
 *       release.
 */
template<std::size_t D>
class Turing_CartesianGrid
{
public:
  /**
   * Creates an instance of a Cartesian 2D grid.
   * @param[in] x Number of cells in x direction.
   * @param[in] y Number of cells in y direction.
   * @param[in] h Size of the grid cells' edges.
   */
  Turing_CartesianGrid (const std::array<std::size_t,D> N, const double h)
    : N_(N), h_(h) {}

  /**
   * Returns the total number of grid cells.
   */
  std::size_t size () const
  {
    return std::accumulate(N_.begin(), N_.end(), 1., std::multiplies<std::size_t>());
  }

  /**
   * Returns the number of grid cells in direction d.
   */
  std::size_t size (const std::size_t d) const
  {
    assert(d < D);
    return N_[d];
  }

  /**
   * Returns the mesh width, i.e. the size of the grid cells' edges.
   */
  double meshwidth () const
  {
    return h_;
  }

  /**
   * Returns the index of a cell based on the cell's multi-index i;
   * this induces a linear ordering of the grid cells which can be
   * used to assign data to the grid like a piecewise constant grid
   * function, i.e. a function with a constant value on each cell.
   */
  std::size_t index (const std::array<std::size_t,D>& i) const
  {
    std::size_t index = 0;
    std::size_t offset = 1;
    for(std::size_t d=0; d<D; ++d)
      {
	index += i[d]*offset;
	offset *= N_[d];
      }
    return index;    
  }

  /**
   * Returns the neighbors of a cell based on the cell's
   * multi-index i; the neighbors are returned as a list of indexes.
   */
  std::pair<std::array<std::size_t,2*D>,std::size_t> neighbours (const std::array<std::size_t,D>& i) const
  {
    std::array<std::size_t,2*D> n = {};
    std::size_t end = 0;

    // copy index of the cell we are interested in
    std::array<std::size_t,D> j = i;

    // get neighbors in each direction d
    for (unsigned int d = 0; d < D; d++)
    {
      // check left (with respect to direction d)
      if (i[d] > 0){
	j[d]--;
	n[end++] = index(j);
	j[d]++;
      }
      // check right (with respect to direction d)
      if (i[d] < N_[d]-1){
	j[d]++;
	n[end++] = index(j);
	j[d]--;
      }
    }
    return {n,end};
  }

private:
  std::array<std::size_t,D> N_;
  double h_;
};


#endif // TURING_CARTESIANGRID_HH
