#ifndef TURING_VTKWRITER_HH
#define TURING_VTKWRITER_HH

#include <cstdio>

#include <string>
#include <vector>
#include <stdexcept>

#include "turing_cartesiangrid.hh"


/**
 * A class for data output in our turing model code.
 * Data is written in the VTK file format which is readable
 * e.g. by ParaView (www.paraview.org).
 */
class Turing_VTKWriter
{
public:
  /**
   * Creates an instance of a VTK writer.
   * @param[in] basename The basic filename for data output.
   * @param[in] grid     Grid which is used in the simulation.
   * @param[in] data     Vector which contains the solution's
   *                     function value in each grid cell.
   */
  Turing_VTKWriter (std::string basename,
                    const Turing_CartesianGrid<2>& grid,
                    const std::vector<double>& data)
    : count_(0), basename_(basename), grid_(grid), data_(data)
  {}

  /**
   * Write a single time step of the simulated turing model.
   * @param[in] c Current file number (for the filename), optional.
   */
  void write (int c = -1)
  {
    if (c == -1)
      c = count_;

    // create filename
    char fname[256];
    snprintf(fname,255,"%s-%05i.vtk", basename_.c_str(), c);

    // create file
    FILE* out = fopen(fname, "w");
    if (!out)
      throw std::runtime_error(std::string("Failed to open file ") + fname);

    // write header
    fprintf(out,"# vtk DataFile Version 2.0\n");
    fprintf(out,"vtk output\n");
    fprintf(out,"BINARY\n ");
    fprintf(out," \n");
    fprintf(out," \n");
    fprintf(out,"DATASET STRUCTURED_POINTS\n");
    fprintf(out,"DIMENSIONS %lu %lu 1\n",
            grid_.size(0)+1, grid_.size(1)+1);
    fprintf(out,"ORIGIN %f %f %f\n", 0.0, 0.0, 0.0);
    fprintf(out,"SPACING %f %f %f\n",
            grid_.meshwidth(), grid_.meshwidth(), grid_.meshwidth());
    fprintf(out," \n");
    fprintf(out,"CELL_DATA %lu\n", grid_.size());
    fprintf(out,"SCALARS %s float 1\n", basename_.c_str());
    fprintf(out,"LOOKUP_TABLE default\n");

    // write data
    std::array<std::size_t,2> i= {};
    for (i[1] = 0; i[1] < grid_.size(1); i[1]++)
      for (i[0] = 0; i[0] < grid_.size(0); i[0]++)
      {
        unsigned int idx = grid_.index(i);

        // get function value in current grid cell
        float f = data_[idx];

        // byte swap
        char* c = reinterpret_cast<char*>(&f);
        static const int n = sizeof(float);
        for (int i = 0; i < n/2; i++)
        {
          char t = c[n-1-i];
          c[n-1-i] = c[i];
          c[i] = t;
        }

        // write function value in current grid cell
        fwrite(&f, 1, n, out);
        if (ferror(out))
        {
          std::string error
            = std::string("Failed to write data to file ") + fname;
          throw std::runtime_error(error);
        }
      }

    // close file
    fclose(out);

    // update timestep
    count_ = c+1;
  }

private:
  unsigned int count_;
  const std::string basename_;
  const Turing_CartesianGrid<2>& grid_;
  const std::vector<double>& data_;
};


#endif // TURING_VTKWRITER_HH
